#!/usr/local/bin/perl -w 

# Sval1to2.pl v0.31

# Program to convert training and test files from Senseval 1 format to Senseval 2 format

# Copyright (C) 2000-2002
# Ted Pedersen, University of Minnesota, Duluth
# tpederse@d.umn.edu
# Satanjeev Banerjee, University of Minnesota, Duluth
# bane0025@d.umn.edu
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
#-----------------------------------------------------------------------------
#                              Start of program
#-----------------------------------------------------------------------------

# this program only works for version 5.6 or higher
require 5.006_000;

# we have to use commandline options, so use the necessary package!
use Getopt::Long;

# first check if no commandline options have been provided... in which case
# print out the usage notes!
if ( $#ARGV == -1 )
{
    &minimalUsageNotes();
    exit;
}

# now get the options!
GetOptions("training=s", "test=s", "sensemap=s", "split_on_POS", "clean_up", "numeric_answers_only", "floating", "help", "version");

# if help has been requested, print out help!
if ( defined $opt_help )
{
    $opt_help = 1;
    &showHelp();
    exit;
}

# if version has been requested, show version!
if ( defined $opt_version )
{
    $opt_version = 1;
    &showVersion();
    exit;
}

if (defined $opt_split_on_POS)
{
    $opt_split_on_POS = 1;
}

# if sensemap has been given, open it and read it, otherwise complain.
if (defined $opt_sensemap)
{
    open(SENSE, $opt_sensemap) || die("Couldnt open $opt_sensemap");
    
    while (<SENSE>)
    {
	my @record = split(/\s+/, $_);
	my $key;

	# ignore the U, T, P records... we'll include them in every file anyways!
	# CHANGE: Jan3, 2003, M.Saif 
	if (defined $record[1]) {
	if ($record[1] =~ /^99999[789]$/) { next; } 

	# create hash with the sense number as key and word-POS as the value
	if (defined $senseHash{$record[1]})
	{
	    	print "sense $key already has a number!\n"; 
	}
	else { $senseHash{$record[1]} = "$record[0]-$record[3]"; } }
    }
}
else
{
    print STDERR "Switch --sensemap is a compulsory switch.\n";
    askHelp();
    exit;
}

# if train directory has been given, open it and read it, otherwise complain.
if (defined $opt_training)
{
    if (!(-d $opt_training))
    {
	print STDERR "$opt_training does not look like a directory.\n";
	askHelp();
	exit;
    }

    opendir DIR, $opt_training || die ("Couldnt open directory $opt_training!\n");
    my @files = grep /.cor$/, readdir DIR;
    closedir DIR;

    while (@files)
    {
	push (@train_files, $opt_training . "/" . (shift @files));
    }
}
else
{
    print STDERR "Switch --training is a compulsory switch.\n";
    askHelp();
    exit;
}

# if we couldnt find even one file, complain and quit!
if (!(@train_files))
{
    print STDERR "No training files found in directory $opt_training!\n";
    askHelp();
    exit;
}

# if test directory has been given, open it and read it, otherwise complain.
if (defined $opt_test)
{
    if (!(-d $opt_test))
    {
	print STDERR "$opt_test does not look like a directory.\n";
	askHelp();
	exit;
    }

    opendir DIR, $opt_test || die ("Couldnt open directory $opt_test!\n");
    my @files = grep /.eval$/, readdir DIR;
    closedir DIR;

    while (@files)
    {
	push (@test_files, $opt_test . "/" . (shift @files));
    }
}
else
{
    print STDERR "Switch --test is a compulsory switch.\n";
    askHelp();
    exit;
}


# if we couldnt find even one file, complain and quit!
if (!(@test_files))
{
    print STDERR "No test files found in directory $opt_test!\n";
    askHelp();
    exit;
}

# first do the easy job: convert the test files
print STDERR "Converting test files...\n";

foreach $source (@test_files)
{
    print STDERR "$source\n";

    # deduce the file type and the lexelt from the file name. we already know the filename
    # has a .eval extension
    if ($source =~ /(\w+-\w)\.eval/)
    {
	$lexelt = $1; 
    }
    else 
    {
	print STDERR "POS information not available in filename $source. Skipping!\n";
	next;
    }

    # save this lexelt for use later by the training half of this program
    $evalLexelts{$lexelt} = 1;

    # open the lexelt.xml file for output
    open (DST, ">$lexelt.test.xml") || die ("Couldnt open $lexelt.test.xml");

    # output the top
    print DST "<corpus lang=\"en\">\n";
    print DST "<lexelt item=\"$lexelt\">\n";

    # now open the source file!
    open SRC, "$source" || die "Couldnt open $source\n";
    
    $line = <SRC>;
    # CHANGE: Jan16, 2003, M.Saif
    # Ignore all initial blank lines
    while($line =~ /^\s*$/)	{ $line = <SRC>; }
    # CHANGE: Jan21, 2003, M.Saif
    # Ignore the first non-blank line if it is `n', `v', `a' or `p'
    if ($line =~ /^\s*[nvap]\s*$/)	{ $line = <SRC>; }

    if (!(defined $line)) { $flag = 1; }
    else { $flag = 0; chomp $line; }

    # CHANGE: Jan3, 2003, M.Saif 
    $blank = 1;    
    while ($flag == 0)
    {
	# keep reading until we come to a line with only a solitary
	# number on it, or we reach the end of the file

	# CHANGE: Jan3, 2003, M.Saif
        # Set instance ID flag
        if($blank == 1) { $idflag = 1; }
        else { $idflag = 0;}
        # Set Blank line flag
        if($line =~ /^\s*$/) { $blank = 1; }
        else { $blank = 0; }
 
# space required in /^\d+ $/ below because TEST/bitter-p.eval has a
# space after 700001 on the first line. bug pointed out and fixed by
# Yoong Keok Lee, leeyoong@comp.nus.edu.sg

 	# while ($flag == 0 && !($line =~ s/^(\d+) *$/$1/))
	while ($flag == 0 && ($idflag == 0))
	{
	    $line = <SRC>;

	    # CHANGE: Jan3, 2003, M.Saif
            # Set instance ID flag
	    if (defined $line)
	    {
            	if($blank == 1) { $idflag = 1; }
            	else { $idflag = 0;}
            	# Set Blank line flag
            	if($line =~ /^\s*$/) { $blank = 1; }
            	else { $blank = 0; }
	    }

	    if (!(defined $line)) { $flag = 1; }
	    else { chomp $line; }
	}
	if ( $flag ) { next; }

	# CHANGE: Jan3, 2003, M.Saif
        $line =~ s/^\s*(.*[^\s])\s*/$1/;
	# so thats our instance id
	my $instanceId = $line;
	
	# now get the context for this instance
	$context = "";
	$line = <SRC>;
	if (!(defined $line)) { $flag = 1; next; }
	# CHANGE: Jan3, 2003, M.Saif
	$blank = 0;
	$idflag = 0;
	# while ( $flag == 0 && !($line =~ /^\d+$/))
	while ( $flag == 0 && ($idflag == 0))
	{
	    $context .= $line;
	    $line = <SRC>;

	    if (defined $line)
	    {
		# CHANGE: Jan3, 2003, M.Saif
	        # Set instance ID flag
            	if($blank == 1) { $idflag = 1; }
            	else { $idflag = 0;}
            	# Set Blank line flag
            	if($line =~ /^\s*$/) { $blank = 1; }
            	else { $blank = 0; }
	    }

	    if (!(defined $line)) { $flag = 1; }
	    else { chomp $line; }
	}
	# CHANGE: Jan3, 2003, M.Saif
	$blank = 1;
	
	# convert from <tag> format to <head> format
	$context =~ s/<tag[^>]*>([^<]*)<\/>/<head>$1<\/head>/;

	# do some low level "clean-up if requested. Following lines retain only letters, 
	# numbers and space. Removes all capitalization, converts multiple white space to 
	# a single white space, and removes new lines. 
	
	if (defined $opt_clean_up)
	{
	    $opt_clean_up = 1;
	    $context =~ s/[^\w\s\d<>\/]//g;
	    $context = lc($context);
	    $context =~ s/\s+/ /g;
	}

	# write to file
	print DST "<instance id=\"$lexelt.$instanceId\">\n";
	print DST "<context>\n$context\n</context>\n";
	print DST "</instance>\n";
    }

    # thats the end of this source file. round it off!
    print DST "</lexelt>\n</corpus>\n";
    close DST;

    # close the source file
    close SRC;
}

# that completes it for the test files. now for the training files
print STDERR "Done with test files. Onto the training files...\n";

foreach $source (@train_files)
{
    print STDERR "$source\n";

    # deduce the file type and the lexelt from the file name. we already know that
    # its a .cor file
    $source =~ /(\w+)\.cor/;
    $word = $1; # thats the word we are dealing with

    # we need the currentFiles hash to be cleared
    %currentFiles = ();

    # now open the source file!
    open SRC, "$source" || die "Couldnt open $source\n";
    
    # reset the "extras" string. this string will contain all the U, T and P instances which we'll append
    # to all the destination files at the end. 
    $extras = ""; 

    $line = <SRC>;
    # CHANGE: Jan16, 2003, M.Saif
    while($line =~ /^\s*$/)     { $line = <SRC>; }
    # CHANGE: Jan21, 2003, M.Saif
    # Ignore the first non-blank line if it is `n', `v', `a' or `p'
    if ($line =~ /^\s*[nvap]\s*$/)      { $line = <SRC>; }

    if (!(defined $line)) { $flag = 1; }
    else { $flag = 0; chomp $line; }
   
    # CHANGE: Jan3, 2003, M.Saif
    $blank = 1; 
    while ($flag == 0)
    {
	# keep reading until we come to a line with only a solitary
	# number on it, or we reach the end of the file
	# CHANGE: Jan3, 2003, M.Saif
	# Set instance ID flag
	if($blank == 1) { $idflag = 1; }
	else { $idflag = 0;}
	# Set Blank line flag
	if($line =~ /^\s*$/) { $blank = 1; }
	else { $blank = 0; }
 
	# while ($flag == 0 && !($line =~ /^\d+$/))
	while ($flag == 0 && ($idflag == 0))
	{
	    $line = <SRC>;
	
	    # CHANGE: Jan3, 2003, M.Saif
            # Set instance ID flag
            if (defined $line)
            {
                if($blank == 1) { $idflag = 1; }
                else { $idflag = 0;}
                # Set Blank line flag
                if($line =~ /^\s*$/) { $blank = 1; }
                else { $blank = 0; }
            }

	    if (!(defined $line)) { $flag = 1; }
	    else { chomp $line; }
	}
	if ( $flag ) { next; }

	# CHANGE: Jan3, 2003, M.Saif
	$line =~ s/^\s*(.*[^\s])\s*/$1/;
	# so thats our instance id
	my $instanceId = $line;
	
	# now get the context for this instance
	$context = "";
	$line = <SRC>;
	if (!(defined $line)) { $flag = 1; next; }
	# CHANGE: Jan3, 2003, M.Saif
	$blank = 0;
	$idflag = 0;
	# while ( $flag == 0 && !($line =~ /^\d+$/))
	while ( $flag == 0 && ($idflag == 0))
	{
	    $context .= $line;
	    $line = <SRC>;

	    if (defined $line)
            {   
                # CHANGE: Jan3, 2003, M.Saif
                # Set instance ID flag
                if($blank == 1) { $idflag = 1; }
                else { $idflag = 0;}
                # Set Blank line flag
                if($line =~ /^\s*$/) { $blank = 1; }
                else { $blank = 0; }
            }   

	    if (!(defined $line)) { $flag = 1; }
	    else { chomp $line; }
	}
	# CHANGE: Jan3, 2003, M.Saif
	$blank = 1;
	
	# take out the answer string, and convert the tags to the <head></head> format 
	if ( $context =~ /<tag\s([^>]*)>/ )
	{
	    $answerString = $1;

	    # change to <head> format
	    $context =~ s/<tag[^>]*>([^<]*)<\/>/<head>$1<\/head>/;
	}

	# do some low level "clean-up if need be. Following lines retain only letters, 
	# numbers and space. Removes all capitalization, converts multiple white space to 
	# a single white space, and removes new lines. 
	
	if (defined $opt_clean_up)
	{
	    $opt_clean_up = 1;
	    $context =~ s/[^\w\s\d<>\/]//g;
	    $context = lc($context);
	    $context =~ s/\s+/ /g;
	}

	# now get hold of the answers and do all sorts of things with them!
	if (defined @answers) { undef @answers; }

	# CHANGE: Jan3, 2003, M.Saif
	# while ($answerString =~ /(\d{6}(-\w)?)/g)
	#{
        #    push @answers, $1;
        #}

	# Clean up the answer string
	$answerString  =~ s/\"//g;
	$answerString  =~ s/^\s*//;
	$answerString  =~ s/\s*$//;
	# Check for multiple answers
	if($answerString  =~ / or /) {
		@answers = split / or /, $answerString; }
	else {
		push @answers, $answerString; }
	
	# if no answer quit! training instances require answers!
	if (!(@answers)) 
	{ 
	    print STDERR "Instance no $instanceId in file $source does not have answers!\n"; 
	    exit;
	}
	
	# now go thru the answers list, and find out the POS of the word. from that, deduce
	# the lexelt and the file name. if file already opened, write to it, otherwise open the file!
	# in case answer is U, T or P, it should go to all the files. since we do not know how many
	# files we have until we have finished going thru the whole source file, just write out
	# all these UTP instances to a single string. at the end of this source file, we shall append 
	# it to all the files! (COMPLICATED AS HELL!!)
	
	foreach $answer (@answers)
	{
	    # for each answer we'll decide a (possibly different) lexelt and filehandle
	    if (defined $lexelt ) { undef $lexelt; }
	    if (defined $fileHandle) { undef $fileHandle; }

	    # CHANGE: Jan3, 2003, M.Saif
	    # Here we want to remove the -p, -a, -m kind of suffixes to the answer ID 
	    # the answer may have a -\w attached to it. get hold of the numeric part
	    # $answer =~ /(\d+)/;
	    # my $numericAns = $1;
	    $answer =~ /^(.*[^-][^?])(-.)?\??$/;
	    my $numericAns = $1;
	    
	    if ($numericAns =~ /^99999[789]/)
	    {
		# ok, UTP answer! write to $extras
		# note, instance id should include the lexelt. but we dont know the lexelt yet, so 
		# we'll fill it in when we put this string in the respective files
		
		$extras .= "<instance id=\"$instanceId\">\n"; 
		
		if (defined $opt_numeric_answers_only)
		{
		    $extras .= "<answer instance=\"$instanceId\" senseid=\"$numericAns\"/>\n";
		}
		else 
		{
		    $extras .= "<answer instance=\"$instanceId\" senseid=\"$answer\"/>\n";
		}
		
		$extras .= "<context>\n$context\n</context>\n</instance>\n";
		next;
	    }
	    
	    # ok so its not a UTP answer. deduce lexelt
	    # if split is not defined or word-p exists, then dont split!

	    if (!defined ($opt_split_on_POS))
	    {
		# ok so we are not splitting files anyways. so the word is the lexelt. great! 
		$lexelt = $word;
	    }

	    elsif (defined ($evalLexelts{"$word-p"}))
	    {
		# so we ARE allowed to split, but we cant for this word thanx to a -p eval file
		# ok, the lexelt is $word-p
		$lexelt = $word . "-p";
	    }

	    else 
	    {
		# hmm so we are splitting. now figure out the lexelt from the sense of the answer
		if (!defined($senseHash{$numericAns}))
		{
		    print STDERR "Couldnt find answer $numericAns of instance $instanceId of file $source in sensemap\n";
		    exit;
		}
	    
		$lexelt = $senseHash{$numericAns};

		if (defined $opt_floating)
		{
		    $opt_floating = 1;
		    if ($lexelt eq "float-a") { $lexelt = "floating-a"; }
		}
	    }
	    
	    # right so that gets us a lexelt. there should be a training file by this lexelt open.
	    # if not, then open one
	    if (!defined $currentFiles{$lexelt})
	    {
		# so open it
		open (my $file, ">$lexelt.train.xml") || die ("Couldnt open file $lexelt.train.xml");
		$currentFiles{$lexelt} = $file;
		
		# print the usual starting stuff
		print $file "<corpus lang=\"en\">\n";
		print $file "<lexelt item=\"$lexelt\">\n";
	    }
	    
	    # get the file handle for this lexelt
	    $fileHandle = $currentFiles{$lexelt};
	    
	    # print the rest of the instance
	    print $fileHandle "<instance id=\"$lexelt.$instanceId\">\n";
	    if (defined $opt_numeric_answers_only)
	    {
		print $fileHandle "<answer instance=\"$lexelt.$instanceId\" senseid=\"$numericAns\"/>\n";
	    }
	    else 
	    {
		print $fileHandle "<answer instance=\"$lexelt.$instanceId\" senseid=\"$answer\"/>\n";
	    }
	    
	    print $fileHandle "<context>\n$context\n</context>\n</instance>\n";
	}
    }
    
    # thats the end of this source file. put in the extras into all the opened training output files
    foreach $lexelt (keys %currentFiles)
    {
	# first put in the lexelt into the instance id and the answer
	my $temp = $extras;
	$temp =~ s/id=\"([^\"]*)\"/id=\"$lexelt.$1\"/g;
	$temp =~ s/answer instance=\"([^\"]*)\"/answer instance=\"$lexelt.$1\"/g;

	# now get the filehandle
	$fileHandle = $currentFiles{$lexelt};

	# now write out temp into the file
	print $fileHandle $temp;
	
	# round off the file, close it and remove the filehandle from the hash
	print $fileHandle "</lexelt>\n</corpus>\n";
	close $fileHandle;
	delete $currentFiles{$lexelt};
    }

    # close the source file
    close SRC;
}

# thats that!

# function to output a minimal usage note when the user has not provided any
# commandline options
sub minimalUsageNotes
{
    print "Usage: Sval1to2.pl [OPTIONS] --training TRAIN_DIR --test TEST_DIR --sensemap MAP\n";
    askHelp();
}

# function to output help messages for this program
sub showHelp
{
    print "Usage: Sval1to2.pl [OPTIONS] --training TRAIN_DIR --test TEST_DIR --sensemap MAP\n\n";

    print "Converts training and test files from Senseval 1 format to Senseval 2 format\n";
    print "format. Training files (extension .cor) should all be in TRAIN_DIR while\n";
    print "test files (extension .eval) should all be in TEST_DIR. Sensemap is the file\n";
    print "mne-uid.map as available in the Senseval 1 data. Output training files\n";
    print "(extension .train.xml) and output test files (extension .test.xml) in Senseval\n";
    print "2 format are created in the current directory.\n\n";

    print "Compulsory Switches:\n\n";

    print "  --training TRAIN_DIR     Directory containing the training files. Training files\n";
    print "                           need to be of extension .cor.\n\n";
    print "  --test TEST_DIR          Directory containing the test files. Test files need to\n";
    print "                           be of extension .eval.\n\n";
    print "  --sensemap MAP           MAP refers to the mne-uid map file that is available with\n";
    print "                           Senseval 1 data.\n\n";

    print "OPTIONS (Optional Switches):\n\n";

    print "  --split_on_POS           Splits a single training file into separate POS files if\n";
    print "                           corresponding POS files are present in the test directory.\n";
    print "                           Eg: Since we have both bet-n.eval and bet-v.eval,\n";
    print "                           instances in bet.cor will be split into bet-n.train.xml and\n";
    print "                           bet-v.train.xml. By default, no POS splitting is done and\n";
    print "                           each .cor training file results in exactly one .train.xml file.\n\n";
    print "  --clean_up               \"Cleans up\" the context. Retains only letters, numbers\n";
    print "                           and space. Removes all capitalization, converts multiple\n";
    print "                           white space to single white space, and removes new lines.\n\n";
    print "  --numeric_answers_only   Retain only numeric portion of answers. Useful in getting\n";
    print "                           rid of the non-numeric (hyphenated) portion of some answers.\n\n";
    print "  --floating               Creates floating-a.train.xml instead of float-a.train.xml.\n";
    print "                           Makes script writing that much easier :-)\n\n";
    print "  --help                   Prints this help message.\n\n";
    print "  --version                Prints the version number.\n\n";
}

# function to output the version number
sub showVersion
{
    print "Sval1to2.pl  -  Version 0.3\n";
    print "Copyright (C) 2000-2002, Ted Pedersen & Satanjeev Banerjee\n";
}

# function to output "ask for help" message when the user's goofed up!
sub askHelp
{
    print STDERR "Type Sval1to2.pl --help for help.\n";
}


