#!/usr/bin/perl -w

# keyOne2Two.pl v0.2
# Program to convert answers from Senseval 1 format to Senseval 2 format

# Copyright (C) 2001-2002
# Ted Pedersen, University of Minnesota, Duluth
# tpederse@d.umn.edu
# Satanjeev Banerjee, University of Minnesota, Duluth
# bane0025@d.umn.edu
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
#-----------------------------------------------------------------------------
#                              Start of program
#-----------------------------------------------------------------------------

# we have to use commandline options, so use the necessary package!
use Getopt::Long;

# first check if no commandline options have been provided... in which case
# print out the usage notes!
if ($#ARGV == -1)
{
    &minimalUsageNotes();
    exit;
}

# now get the options!
GetOptions("version", "help", "sensemap=s");

# if help has been requested, print out help!
if (defined $opt_help)
{
    $opt_help = 1;
    &showHelp();
    exit;
}

# if version has been requested, show version!
if (defined $opt_version)
{
    $opt_version = 1;
    &showVersion();
    exit;
}

# if sensemap has been given, open it and read it!
if (defined $opt_sensemap)
{
    open(SENSE, $opt_sensemap) || die("Couldnt open $opt_sensemap");
    
    while (<SENSE>)
    {
	my @record = split(/\s+/, $_);
	my $key;

	if ($record[2] =~ /^[UTP]$/) {$key = $record[2];}
	else { $key = $record[0] . "-" . $record[2]; }
	
	if (defined $senseHash{$key})
	{
	    print "sense $key already has a number!\n";
	    # exit;
	}
	else { $senseHash{$key} = $record[1]; }
    }
}

# now to get hold of the GOLD files specified on the commandline
# now get hold of all the files / directories on the command line
my @commandLineSources = @ARGV;

# now go thru each file checking if it exists. if not, remove that
# file from the list. also check if any of the sources is actually a
# directory. if so, include all files in that directory

my @sources = ();
my $source;

foreach $source (@commandLineSources)
{
    # see if it exists even!
    if (!(-e $source))
    {
	next; 
    }

    # remove the / in the end if provided!
    while ($source =~ /\/$/) { $source =~ s/\/$//; }

    # now see if its a directory
    if (-d $source)
    {
	opendir DIR, $source || "Couldnt open directory $source!\n";
	my @files = grep !/^\./, readdir DIR;
	closedir DIR;

	while (@files)
	{
	    push (@sources, $source . "/" . (shift @files));
	}
	next;
    }

    # so its a file. push.
    push (@sources, $source);
}

# if we couldnt find even one file, complain and quit!
if ($#sources < 0)
{
    print STDERR "No source file specified!\n";
    askHelp();
    exit(-1);
}

# now for the conversion!
foreach $source (@sources)
{
    # get the "lexelt" 
    $source =~ /(\w+-\w)$/;
    my $lexelt = $1;

    # get the stub before the part of speech
    my $stub;
    if ($lexelt =~ /floating/) { $stub = "float"; }
    else 
    {
	$lexelt =~ /^(\w+)/;
	$stub = $1;
    }

    # open the source file!
    open(SOURCE, $source) || die("Couldnt open file $source\n");

    while(<SOURCE>)
    {
	my @record = split(/:/, $_);
	print "$lexelt $lexelt.$record[0] ";
	
	my @senses = split(/ or /, $record[1]);
	while (@senses)
	{
	    my $sense = $senses[0];
	    chomp $sense;
	    
	    if (defined $opt_sensemap)
	    {
		my $key;
		if ($sense =~ /^[UTP]$/) { $key = $sense; }
		else { $key = $stub . "-" . $sense; }

		if (!(defined $senseHash{$key}))
		{
		    print STDERR "Couldnt find sense mapping for sense $sense of file $lexelt, instance $record[0]\n";
		}
		else 
		{
		    print "$senseHash{$key} ";
		}
	    }
	    else { print "$sense "; }
	    shift @senses;
	}
	print "\n";
    }
    close SOURCE;
}

# thats that!

# function to output a minimal usage note when the user has not provided any
# commandline options
sub minimalUsageNotes
{
    print "Usage: keyOne2Two.pl [OPTIONS] SOURCE [SOURCE...]\n";
    askHelp();
}

# function to output help messages for this program
sub showHelp
{
    print "Usage: keyOne2Two.pl [OPTIONS] SOURCE [SOURCE...]\n\n";
    print "Converts \"GOLD\" files of Senseval 1 format to key file in Senseval 2 format. SOURCE\n";
    print "should be either a> GOLD file in Senseval 1 format, or b> directory containing GOLD\n";
    print "files in Senseval 1 format, or a mixture of such files and directories. The\n";
    print "corresponding Senseval 2 formatted key file is written to stdout. If the sense-map file\n";
    print "(mne-uid.map) is provided, then mnemonics used in the GOLD file are converted to numbers\n";
    print "according to the sense-map file. The resulting key file is output to stdout.\n\n";

    print "OPTIONS:\n\n";
    print "  --sensemap         Mappings from mnemonics to sense-numbers are picked up from\n";
    print "                     this file.\n\n";
    print "  --version          Prints the version number.\n\n";
    print "  --help             Prints this help message.\n\n";
}

# function to output the version number
sub showVersion
{
    print "keyOne2Two.pl  -  Version 0.2\n";
    print "Copyright (C) 2001-2002, Ted Pedersen & Satanjeev Banerjee\n";
}

# function to output "ask for help" message when the user's goofed up!
sub askHelp
{
    print "Type keyOne2Two.pl --help for help.\n";
}
